@@ -16,7 +16,7 @@ from notte_core.actions import (
 )
 from notte_core.agent_types import AgentCompletion
 from notte_core.browser.observation import ExecutionResult, Observation, TrajectoryProgress
-from notte_core.common.config import NotteConfig, RaiseCondition
+from notte_core.common.config import LlmModel, NotteConfig, RaiseCondition
 from notte_core.common.telemetry import track_usage
 from notte_core.common.tracer import LlmUsageDictTracer
 from notte_core.credentials.base import BaseVault, LocatorAttributes
@@ -126,10 +126,12 @@ class NotteAgent(BaseAgent):
 
         with ErrorConfig.message_mode("developer"):
             response: AgentCompletion = await self.llm.structured_completion(
-                messages, response_format=AgentCompletion, use_strict_response_format=False
+                messages,
+                response_format=AgentCompletion,
+                use_strict_response_format=LlmModel.use_strict_response_format(self.config.reasoning_model),
             )
 
-        self.trajectory.append(response, force=True)
+        await self.trajectory.append(response, force=True)
         return response
 
     @profiler.profiled()
@@ -158,7 +160,7 @@ class NotteAgent(BaseAgent):
                 # Human in the loop is not implemented yet => fail immediately
                 error_msg = f"Agent requested human help: {reason}"
                 ex_res = ExecutionResult(action=response.action, success=False, message=error_msg)
-                self.trajectory.append(ex_res, force=True)
+                await self.trajectory.append(ex_res, force=True)
                 return CompletionAction(success=False, answer=error_msg)
             case CaptchaSolveAction(captcha_type=captcha_type) if (
                 not self.session.window.resource.options.solve_captchas
@@ -166,13 +168,13 @@ class NotteAgent(BaseAgent):
                 # if the session doesnt solve captchas => fail immediately
                 error_msg = f"Agent encountered {captcha_type} captcha but session doesnt solve captchas: create a session with solve_captchas=True"
                 ex_res = ExecutionResult(action=response.action, success=False, message=error_msg)
-                self.trajectory.append(ex_res, force=True)
+                await self.trajectory.append(ex_res, force=True)
                 return CompletionAction(success=False, answer=error_msg)
 
             case CompletionAction(success=False, answer=answer):
                 # agent decided to stop with failure
                 result = ExecutionResult(action=response.action, success=False, message=answer)
-                self.trajectory.append(result, force=True)
+                await self.trajectory.append(result, force=True)
                 return response.action
             case CompletionAction(success=True, answer=answer) as output:
                 # need to validate the agent output
@@ -189,7 +191,7 @@ class NotteAgent(BaseAgent):
                     logger.info(f"Validation successful: {val_result.message}")
                     logger.info("✅ Task completed successfully")
                     result = ExecutionResult(action=response.action, success=True, message=val_result.message)
-                    self.trajectory.append(result, force=True)
+                    await self.trajectory.append(result, force=True)
                     # agent and validator agree, stop with success
                     return response.action
 
@@ -200,7 +202,7 @@ class NotteAgent(BaseAgent):
                     "perform actions that would prove it is."
                 )
                 result = ExecutionResult(action=response.action, success=False, message=agent_failure_msg)
-                self.trajectory.append(result, force=True)
+                await self.trajectory.append(result, force=True)
                 # disagreement between model and validation, continue
                 return None
             case _:
@@ -227,6 +229,7 @@ class NotteAgent(BaseAgent):
         return TrajectoryProgress(current_step=self.trajectory.num_steps, max_steps=self.config.max_steps)
 
     @track_usage("local.agent.messages.get")
+    @profiler.profiled()
     async def get_messages(self, task: str) -> list[AllMessageValues]:
         """
         Formats a trajectory into a list of messages for the LLM, including the current observation.
@@ -269,16 +272,20 @@ class NotteAgent(BaseAgent):
                     conv.add_user_message(
                         content=self.perception.perceive_action_result(step, include_ids=False, include_data=True)
                     )
-                case Observation():
+
+                # observation or screenshot
+                case _:
                     # TODO: add partial info for previous?
                     pass
 
         # Add current observation (only if it's not empty)
         last_obs = self.trajectory.last_observation
         if last_obs is not None and last_obs is not Observation.empty():
+            perceived_content = self.perception.perceive(obs=last_obs, progress=self.progress)
+            image = last_obs.screenshot.bytes() if self.config.use_vision else None
             conv.add_user_message(
-                content=self.perception.perceive(obs=last_obs, progress=self.progress),
-                image=(last_obs.screenshot.bytes() if self.config.use_vision else None),
+                content=perceived_content,
+                image=image,
             )
             conv.add_user_message(self.prompt.select_action())
 
@@ -313,7 +320,7 @@ class NotteAgent(BaseAgent):
             raise e
         finally:
             # in case we failed in step, stop it (relevant for session)
-            _ = self.trajectory.stop_step(ignore_not_in_step=True)
+            _ = await self.trajectory.stop_step(ignore_not_in_step=True)
             _ = self.trajectory.stop()
 
     async def _run(self, request: AgentRunRequest) -> AgentResponse:
@@ -328,20 +335,20 @@ class NotteAgent(BaseAgent):
 
         # initial goto, don't do an llm call just for accessing the first page
         if request.url is not None:
-            _ = self.trajectory.start_step()
+            _ = await self.trajectory.start_step()
             _ = await self.session.aobserve(perception_type=self.perception.perception_type)
-            self.trajectory.append(AgentCompletion.initial(request.url), force=True)
+            await self.trajectory.append(AgentCompletion.initial(request.url), force=True)
             _ = await self.session.aexecute(GotoAction(url=request.url))
-            _ = self.trajectory.stop_step()
+            _ = await self.trajectory.stop_step()
 
         step = 0
         while self.trajectory.num_steps < self.config.max_steps:
             step += 1
             logger.info(f"💡 Step {step}")
 
-            _ = self.trajectory.start_step()
+            _ = await self.trajectory.start_step()
             completion_action = await self.step(request)
-            _ = self.trajectory.stop_step()
+            _ = await self.trajectory.stop_step()
 
             if completion_action is not None:
                 return await self.output(request.task, completion_action.answer, completion_action.success)
