@@ -1,8 +1,10 @@
 import typing
 
 from notte_browser.session import NotteSession
+from notte_browser.tools.base import BaseTool
 from notte_core.common.config import NotteConfig
 from notte_core.credentials.base import BaseVault
+from notte_core.trajectory import Trajectory
 from notte_sdk.types import AgentCreateRequest, AgentCreateRequestDict
 
 from notte_agent.agent import NotteAgent
@@ -15,15 +17,17 @@ class GufoAgent(NotteAgent):
         self,
         session: NotteSession,
         vault: BaseVault | None = None,
+        tools: list[BaseTool] | None = None,
+        trajectory: Trajectory | None = None,
         **data: typing.Unpack[AgentCreateRequestDict],
     ):
         _ = AgentCreateRequest.model_validate(data)
         config: NotteConfig = NotteConfig.from_toml(**data)
         super().__init__(
-            prompt=GufoPrompt(),
+            prompt=GufoPrompt(tools=tools),
             perception=GufoPerception(),
             config=config,
             session=session,
-            trajectory=session.trajectory.view(),
+            trajectory=trajectory,
             vault=vault,
         )
